#include "bioloid_dyn_master_servos.h"
#include "bioloid_time.h"
#include "usart1.h"
#include "gpio.h"

#define ENABLE_RX_EN_GPIO_CLK      __GPIOA_CLK_ENABLE()
#define RX_EN_PIN                  GPIO_PIN_8
#define RX_EN_GPIO_PORT            GPIOA                       

#define ENABLE_TX_EN_GPIO_CLK      __GPIOC_CLK_ENABLE()
#define TX_EN_PIN                  GPIO_PIN_9               
#define TX_EN_GPIO_PORT            GPIOC               

#define POWER_GPIO_CLK             __GPIOB_CLK_ENABLE()
#define POWER_PIN                  GPIO_PIN_12
#define POWER_GPIO_PORT            GPIOB                       

/* private variables */
TDynamixelMaster bioloid_dyn_master_servos;
TTime bioloid_dyn_master_servos_timer;
TComm bioloid_dyn_master_servos_comm;

// private functions
void bioloid_dyn_master_servos_set_rx_mode(void)
{
  HAL_GPIO_WritePin(TX_EN_GPIO_PORT,TX_EN_PIN,GPIO_PIN_RESET);
  HAL_GPIO_WritePin(RX_EN_GPIO_PORT,RX_EN_PIN,GPIO_PIN_SET);
}

void bioloid_dyn_master_servos_set_tx_mode(void)
{
  HAL_GPIO_WritePin(RX_EN_GPIO_PORT,RX_EN_PIN,GPIO_PIN_RESET);
  HAL_GPIO_WritePin(TX_EN_GPIO_PORT,TX_EN_PIN,GPIO_PIN_SET);
}

// public functions
void bioloid_dyn_master_servos_init(void)
{
  GPIO_InitTypeDef GPIO_InitStructure;
  TUSART_IRQ_Priorities priorities;
  UART_InitTypeDef Init;

  // initialize timer structure
  time_init(&bioloid_dyn_master_servos_timer,bioloid_time_get_counts_per_us(),bioloid_time_get_counts);

  // initialize GPIO
  ENABLE_RX_EN_GPIO_CLK;
  ENABLE_TX_EN_GPIO_CLK;

  GPIO_InitStructure.Pin = RX_EN_PIN;
  GPIO_InitStructure.Mode = GPIO_MODE_OUTPUT_PP;
  GPIO_InitStructure.Speed = GPIO_SPEED_HIGH;
  GPIO_InitStructure.Pull = GPIO_NOPULL;
  HAL_GPIO_Init(RX_EN_GPIO_PORT, &GPIO_InitStructure);

  GPIO_InitStructure.Pin = TX_EN_PIN;
  HAL_GPIO_Init(TX_EN_GPIO_PORT, &GPIO_InitStructure);

  bioloid_dyn_master_servos_set_rx_mode();

  GPIO_InitStructure.Pin = POWER_PIN;
  GPIO_InitStructure.Mode = GPIO_MODE_OUTPUT_PP;
  GPIO_InitStructure.Speed = GPIO_SPEED_HIGH;
  HAL_GPIO_Init(POWER_GPIO_PORT, &GPIO_InitStructure);

  bioloid_dyn_master_servos_disable_power();

  /* initialize the comm object */
  comm_init(&bioloid_dyn_master_servos_comm,0x01,&bioloid_dyn_master_servos_timer);
  Init.BaudRate     = 1000000;
  Init.WordLength   = UART_WORDLENGTH_8B;
  Init.StopBits     = UART_STOPBITS_1;
  Init.Parity       = UART_PARITY_NONE;
  Init.Mode         = UART_MODE_TX_RX;
  Init.HwFlowCtl    = UART_HWCONTROL_NONE;
  Init.OverSampling = UART_OVERSAMPLING_16;
  
  priorities.irq_priority=0;
  priorities.irq_subpriority=1;
  priorities.dma_rx_priority=1;
  priorities.dma_rx_subpriority=3;
  priorities.dma_tx_priority=1;
  priorities.dma_tx_subpriority=2;

  usart1_init(&bioloid_dyn_master_servos_comm,&Init,&priorities);
  dyn_master_init(&bioloid_dyn_master_servos,&bioloid_dyn_master_servos_comm);
  bioloid_dyn_master_servos.set_rx_mode=bioloid_dyn_master_servos_set_rx_mode;
  bioloid_dyn_master_servos.set_tx_mode=bioloid_dyn_master_servos_set_tx_mode;

  /* configure dynamixel master module */
  dyn_master_set_rx_timeout(&bioloid_dyn_master_servos,50);
  dyn_master_set_return_level(&bioloid_dyn_master_servos,return_all);
}

inline void bioloid_dyn_master_servos_enable_power(void)
{
  HAL_GPIO_WritePin(POWER_GPIO_PORT,POWER_PIN,GPIO_PIN_SET);
}

inline void bioloid_dyn_master_servos_disable_power(void)
{
  HAL_GPIO_WritePin(POWER_GPIO_PORT,POWER_PIN,GPIO_PIN_RESET);
}

